package com.officetetsu.puzzle;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Arrays;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.os.Bundle;
import android.view.Display;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.WindowManager;
import android.view.animation.Animation;
import android.view.animation.AnimationUtils;
import android.widget.Button;
import android.widget.TableRow;
import android.widget.TextView;

public class Puzzle extends Activity 
	implements
	OnClickListener,
	DialogInterface.OnClickListener{
	
	private Button button_bgn, button_fin;
	private TextView textView;
	private AlertDialog complete_dialog;
	private ImagePiece[] imagePiece 
		= new ImagePiece[16];
	private Bitmap[] piece_bitmap 
		= new Bitmap[16];
	private Bitmap piece_bitmap_nouse;
	private Bitmap blank_bitmap;
	private int blank_num;
	private int turn;
	private int board_size;
	private Animation landing;
	//Standard edition only
	private Score[] score = new Score[11];
	//For bug fix
	private int[] directions = {-4, -1, 1, 4};
	private int difficulty = 128;

	
	/** Called when the activity is first created. */
	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		setContentView(R.layout.main);
		
		int[] tableRowId = {
			R.id.tableRow1,
			R.id.tableRow2,
			R.id.tableRow3,
			R.id.tableRow4
		};
		
		for(int row = 0; row < 4; row++){
			TableRow tableRow = (TableRow)findViewById(tableRowId[row]);
			for(int col = 0; col < 4; col++){
				int num = row * 4 + col;
				imagePiece[num] = new ImagePiece(this);
				imagePiece[num].pieceNumber = num;
				imagePiece[num].setPadding(1, 1, 1, 1);
				imagePiece[num].setOnClickListener(this);
				tableRow.addView(imagePiece[num]);
			}
		}
		
		button_bgn = (Button)findViewById(R.id.button1);
		button_bgn.setOnClickListener(this);
		button_fin = (Button)findViewById(R.id.button2);
		button_fin.setOnClickListener(this);
		textView = (TextView)findViewById(R.id.textView1);
		landing = AnimationUtils.loadAnimation(this, R.anim.bounce);

		WindowManager windowmanager = (WindowManager)getSystemService(WINDOW_SERVICE);
		Display display = windowmanager.getDefaultDisplay();
		board_size = display.getWidth();

		Bitmap bitmap = BitmapFactory.decodeResource(getResources(),R.drawable.blank_photo);
		blank_bitmap = Bitmap.createScaledBitmap(
			bitmap,
			board_size / 4 - 2,
			board_size / 4 - 2,
			true);

		attachImage();

		//Standard edition only
		for(int i = 0; i < 11; i++) score[i] = new Score(this);
	}
	
	private void attachImage(){
		turn = 0;
		textView.setText("0");
		button_bgn.setEnabled(true);

		Bitmap bitmap = BitmapFactory.decodeFile(
				getFilesDir().getAbsolutePath() + "/def_photo.jpg");
		if(bitmap == null){
			bitmap = BitmapFactory.decodeResource(
					getResources(),
					R.drawable.def_photo);
		}

		int bitmap_width = bitmap.getWidth();
		int bitmap_height = bitmap.getHeight();

		if(bitmap_width > bitmap_height){
			int trim_size = (bitmap_width - bitmap_height) / 2;
			bitmap = Bitmap.createBitmap(bitmap, trim_size, 0, bitmap_height, bitmap_height);
		} else {
			int trim_size = (bitmap_height - bitmap_width) / 2;
			bitmap = Bitmap.createBitmap(bitmap, 0, trim_size, bitmap_width, bitmap_width);
		}

		bitmap = Bitmap.createScaledBitmap(
				bitmap,
				board_size,
				board_size,
				true);

		for(int row = 0; row < 4; row++){
			int offset_y = (board_size / 4) * row;
			for(int col = 0; col < 4; col++){
				int offset_x = (board_size / 4) * col;
				int num = row * 4 + col;
				piece_bitmap[num] = Bitmap.createBitmap(
						bitmap,
						offset_x + 1, offset_y + 1,
						board_size / 4 - 2, board_size / 4 - 2);
				imagePiece[num].setImageBitmap(piece_bitmap[num]);
				imagePiece[num].bitmapNumber = num;
			}
		}
		blank_num = 0;
		piece_bitmap_nouse = piece_bitmap[blank_num];
	}
	
	//For bug fix
	private void shuffle(){
		int new_location;
		int blank_location = blank_num;

		for(int i = 0; i < difficulty; i++){
			do{
			new_location = blank_location + directions[(int)(Math.random() * 4)];
			}while((new_location < 0) || (new_location > 15));
			
			swapPiece(new_location);
			blank_location = new_location;
		}
	}
	
	@Override
	public void onClick(View arg0) {
		// TODO Auto-generated method stub
		if (arg0 == button_fin) {
			finish();
		}
		else
		if(arg0 == button_bgn){
			blank_num = (int)(Math.random() * 16);

			piece_bitmap_nouse = piece_bitmap[blank_num];
			piece_bitmap[blank_num] = blank_bitmap;
			imagePiece[blank_num].setImageBitmap(piece_bitmap[blank_num]);

			//For bug fix
			shuffle();
//			for(int i = 0; i < 16; i++){
//				swapImage(i, (int)(Math.random() * 16));
//			}
			
			button_bgn.setEnabled(false);
		}
		else {
			int num = ((ImagePiece)arg0).pieceNumber;
			if(swapPiece(num)){
				textView.setText(Integer.toString(++turn));
				boolean complete = false;
				for(int i = 0;
					(i < 16) && (complete = imagePiece[i].isMatch());
					i++);

				if(complete){
					CharSequence judge_message =
						getString(R.string.label_turn) +
						Integer.toString(turn) +
						getString(R.string.message01) +
						"\n";

					if(turn < 500)
						judge_message = judge_message + 
						getString(R.string.judge01);
					else if(turn < 600)
						judge_message = judge_message + 
						getString(R.string.judge02);
					else
						judge_message = judge_message + 
						getString(R.string.judge03);
					
					complete_dialog = new AlertDialog.Builder(this)
					.setTitle(R.string.title_complete)
					.setMessage(judge_message)
					.setPositiveButton("OK", this)
					.show();
					
					//Standard edition only
					readScore();
					score[10].setDateTime();
					score[10].turn = turn;
					Arrays.sort(score, new ScoreComparator());
					writeScore();
				}
			}
		}
	}
	
	//Standard edition only
	private void readScore(){
		try{
			int i;
			byte[] buffer = new byte[24];

			FileInputStream input = openFileInput("score.bin");
			int num = input.available() / 24;
			if(num > 10) num = 10;
			for(i = 0; i < num; i++){
				for(int j = 0; j < 24; j++)
					buffer[j] = (byte)input.read();
				score[i].setBytes(buffer);
			}
			input.close();

		}
		catch (FileNotFoundException e){}
		catch (IOException e){}
	}

	//Standard edition only
	private boolean writeScore(){
		byte[] buffer = new byte[24];

		try{
			FileOutputStream output =
				openFileOutput("score.bin", Context.MODE_PRIVATE);
			for(int i = 0; i < 10; i++){
				if(score[i].turn < Integer.MAX_VALUE){
					score[i].getBytes(buffer);
					output.write(buffer);
				}
			}
			output.close();
		}
		catch (IOException e){
			return false;
		}
		return true;
	}

	@Override
	public void onClick(DialogInterface arg0, int arg1) {
		// TODO Auto-generated method stub
		if(arg0 == complete_dialog)
			resetImage();
	}
	
	private void resetImage(){
		piece_bitmap[blank_num] = piece_bitmap_nouse;
		imagePiece[blank_num].setImageBitmap(piece_bitmap[blank_num]);
		for(int i = 0; i < 16; i++){
			imagePiece[i].setImageBitmap(piece_bitmap[i]);
			imagePiece[i].bitmapNumber = i;
		}
		blank_num = 0;
		piece_bitmap_nouse = piece_bitmap[blank_num];
		turn = 0;
		textView.setText("0");
		button_bgn.setEnabled(true);
	}

	private void swapImage(int num1, int num2){
		int tmp1 = imagePiece[num1].bitmapNumber;
		int tmp2 = imagePiece[num2].bitmapNumber;
		imagePiece[num1].setImageBitmap(piece_bitmap[tmp2]);
		imagePiece[num2].setImageBitmap(piece_bitmap[tmp1]);
		imagePiece[num1].bitmapNumber = tmp2;
		imagePiece[num2].bitmapNumber = tmp1;
		
		imagePiece[num2].startAnimation(landing);
	}

	private boolean swapPiece(int num){

		if(	imagePiece[num].bitmapNumber == blank_num)
			return false;
		if(	(num >= 4) &&
			(imagePiece[num - 4].bitmapNumber == blank_num)){
			swapImage(num, num - 4);
			return true;
		}
		if(	(num < 12) &&
			(imagePiece[num + 4].bitmapNumber == blank_num)){
			swapImage(num, num + 4);
			return true;
		}
		if(	((num % 4) != 0) &&
			(imagePiece[num - 1].bitmapNumber == blank_num)){
			swapImage(num, num - 1);
			return true;
		}
		if(	(((num + 1) % 4) != 0) &&
			(imagePiece[num + 1].bitmapNumber == blank_num)){
			swapImage(num, num + 1);
			return true;
		}
		return false;
	}
	
	public boolean onCreateOptionsMenu(Menu menu){
		super.onCreateOptionsMenu(menu);
		
		getMenuInflater().inflate(R.menu.main_menu, menu);
		return true;
	}
	
	public boolean onPrepareOptionsMenu(Menu menu){
		super.onCreateOptionsMenu(menu);
		menu.findItem(R.id.item1).setEnabled(!button_bgn.isEnabled());
		menu.findItem(R.id.item3).setEnabled(button_bgn.isEnabled());
		menu.findItem(R.id.item4).setEnabled(button_bgn.isEnabled());
		return true;		
	}
	
	public boolean onOptionsItemSelected(MenuItem item){
		super.onOptionsItemSelected(item);
		switch(item.getItemId()){
		case R.id.item1:
			resetImage();
			break;
		case R.id.item2:
			//Standard edition only
			startActivity(new Intent(this, Lanking.class));
			break;
		case R.id.item3:
			Intent intent = new Intent(this, PhotoBrowser.class);
			startActivityForResult(intent, 0);
			break;
		case R.id.item4:
			blank_num = 6;

			piece_bitmap_nouse = piece_bitmap[blank_num];
			piece_bitmap[blank_num] = blank_bitmap;
			imagePiece[blank_num].setImageBitmap(piece_bitmap[blank_num]);
			
			swapImage(5, 6);
			swapImage(9, 5);
			
			turn = 1000;
			textView.setText("1000");
			button_bgn.setEnabled(false);
			break;
		}
		return true;
	}
	
	@Override
	protected void onActivityResult(int requestCode, int resultCode, Intent intent){
		if(requestCode == 0 && resultCode == Activity.RESULT_OK){
			attachImage();
		}
	}
}